
"""

   m can be
   
       n-1/2 for clarinets
       n for most instruments 
   
   for n integer

"""

import math

SPEED_OF_SOUND = 346.1   # 25C
                         # 30C is 349.0

semitone = {
   'C' : 0,
   'D' : 2,
   'E' : 4,
   'F' : 5,
   'G' : 7,
   'A' : 9,
   'B' : 11
}

def fqc(note):
    if not isinstance(note, str): return note
    
    s = semitone[note[0]]
    note = note[1:]
    if note.startswith('b'):
        s -= 1
        note = note[1:]
    if note.startswith('#'):
        s += 1
        note = note[1:]
    
    s += 12 * int(note)
    
    return 440.0 * 2**((s-57)/12.0)



def cot(x):
    return math.cos(x) / math.sin(x)

def area_of_hole_of_diameter(x):
    r = x / 2.0
    return math.pi * (r*r)

def length_of_fqc(fqc, m):
    return SPEED_OF_SOUND / fqc * m * 0.5

def fqc_of_length(length, m):
    return SPEED_OF_SOUND / length * m * 0.5

def end_flange_length_correction(outer_diameter, inner_diameter):
    a = inner_diameter / 2.0
    w = (outer_diameter - inner_diameter) / 2.0
    
    return a * (0.821 - 0.13*((0.42+w/a)**-0.54))

def hole_length_correction(hole_diameter, bore_diameter):
    # As per p.63-64
    outer_correction = 0.7
    inner_correction = 1.3 - 0.9 * hole_diameter/bore_diameter
    a = hole_diameter / 2.0

    result = a * (inner_correction+outer_correction)
    return result

# def closed_hole_length_correction(hole_diameter, bore_diameter, wall_thickness):
#     """ From flute spreadsheet. 
#         Tiny. """
#    return 0.25 * wall_thickness * (hole_diameter/bore_diameter)**2


def func(left_length, right_length, hole_length, substitution_length,
         bore_area, hole_area,
         m):
    """ Solve for zero by varying substitution_length """
    
    k = m * math.pi / substitution_length
    
    return (
        bore_area * cot(k*left_length - k*substitution_length) +
        bore_area * cot(k*right_length) +
        hole_area * cot(k*hole_length)
    )

def length_with_hole(left_length, right_length, hole_length,
                     bore_area, hole_area, m):
    f = lambda x: func(left_length, right_length, hole_length, x,
                       bore_area, hole_area, m)

    # *Cautiously* look for a solution near the end of the pipe        
    step = 1e-4   #Initially step by 0.1mm

    #Initial guess
    x = left_length+right_length
    assert f(x) > 0.0, 'Something is weird, are your inputs sane?'
    
    while step > 1e-10:        
        while f(x) > 0.0:
            x -= step
        x += step    
        step /= 2.0

    return x


def position_hole(desired_length, length, hole_length, bore_area, hole_area, m):
    """ Returns distance from left hand end of tube """

    x = desired_length
    for i in xrange(100):
        error = length_with_hole(x, length-x, hole_length, bore_area, hole_area, m) \
                - desired_length
        x -= error
    
    assert error < 1e-5, 'Hole positioning failed to converge'
    
    return x



def design_instrument(
        base_note,
        base_m,
        outer_diameter_mm,
        inner_diameter_mm,
        hole_spec):
    """
        Place the finger holes of a cylindrical instrument.
        
        hole spec is a list of tuples defining the holes from the end inward. 
        The tuples consist of:
            the desired note
            the m of the desired note
            the diameter of the hole in mm
            a list indicating which of the previous holes are closed when determining the hole position
        
        What is the m value?
        
            Instruments with a mouthpiece that behaves like a closed end of a pipe (reeds, lips),
            should use an m value of 0.5 (1.5, 2.5, 3.5... for higher registers).
        
            Instruments with a mouthpiece that behaves like an open end of a pipe (flutes, recorders),
            should use an m value of 1 (2, 3, 4... for higher registers).
        
    """
    
    base_fqc = fqc(base_note)
    
    outer_diameter = outer_diameter_mm / 1000.0
    inner_diameter = inner_diameter_mm / 1000.0
    
    wall_thickness = (outer_diameter-inner_diameter)*0.5
    
    base_length = length_of_fqc(base_fqc, base_m)
    end_correction = end_flange_length_correction(outer_diameter, inner_diameter)
    
    print 'Wall thickness    %8.1fmm' % (wall_thickness * 1000.0)
    print 'End correction    %8.1fmm' % (end_correction * 1000.0)
    print
    print 'Instrument length %8.1fmm' % ((base_length-end_correction)*1000.0)
    print
    print 'Note: instrument length includes an end correction but not a mouthpiece length correction'
    print ' - make a pipe about this length, then trim it based on measured frequency'
    print ' - hole positions below are measured from the *end* of the pipe'
    
    print
    print '    Note  Frequency    Hole     Distance   Inter-hole'
    print '                     diameter   from end     spacing'
    print

    prev_pos = None
    holes_so_far = [ ]
        
    for desired_note, m, hole_diameter_mm, holes_closed in hole_spec:
        hole_diameter = hole_diameter_mm / 1000.0
    
        desired_fqc = fqc(desired_note)
        desired_length = length_of_fqc(desired_fqc, m)
        
        assert len(holes_closed) == len(holes_so_far)
        
        current_length = base_length
        for i in xrange(len(holes_so_far)):
            #if holes_closed[i]:
            #    correction = closed_hole_length_correction(
            #        holes_so_far[i][1], 
            #        inner_diameter,
            #        wall_thickness)
            #    #print 'closed hole correction', correction #Tiny
            #    current_length += correction
            if not holes_closed[i]:
                current_length = length_with_hole(
                    holes_so_far[i][0],
                    current_length - holes_so_far[i][0],
                    wall_thickness +
                        hole_length_correction(holes_so_far[i][1], inner_diameter),
                    area_of_hole_of_diameter(inner_diameter),
                    area_of_hole_of_diameter(holes_so_far[i][1]),
                    m)
        
        hole_position = position_hole(
            desired_length, 
            current_length,
            wall_thickness + 
                hole_length_correction(hole_diameter, inner_diameter),  
            area_of_hole_of_diameter(inner_diameter),
            area_of_hole_of_diameter(hole_diameter),
            m)
        
        holes_so_far.append( (hole_position, hole_diameter) )
        
        dist_from_end = base_length-end_correction - hole_position
    
        if prev_pos is not None:
            print ' '*43, '%6.1fmm' % ((dist_from_end-prev_pos)*1000.0)
        prev_pos = dist_from_end
    
        print '%8s %7.1fHz %7.1fmm %8.1fmm' % (desired_note, desired_fqc, hole_diameter*1000.0, dist_from_end*1000.0)
    
        current_length = desired_length
    
    print




